import 'dart:async';
import 'dart:ui';

import 'package:embrace/embrace.dart';
import 'package:embrace_platform_interface/embrace_platform_interface.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mocktail/mocktail.dart';
import 'package:plugin_platform_interface/plugin_platform_interface.dart';

/// PlatformDispatcher.onError is available only for Flutter 3.1 and above.
/// As there is no way to know the Flutter version that runs the test,
/// we enable a flag to disable/enable these tests depending on the version
///
/// To run these test in a version below 3.1 run the following command:
///  `flutter test --dart-define=belowFlutter_3_1=true`
const belowFlutter_3_1 = bool.fromEnvironment('belowFlutter_3_1');

class MockEmbracePlatform extends Mock
    with MockPlatformInterfaceMixin
    implements EmbracePlatform {}

class MockEmbrace extends Mock implements Embrace {}

const errorMessage = '__error__';

class MockError {
  @override
  String toString() => errorMessage;
}

void main() {
  TestWidgetsFlutterBinding.ensureInitialized();

  group('Embrace', () {
    late EmbracePlatform embracePlatform;

    setUp(() {
      embracePlatform = MockEmbracePlatform();
      EmbracePlatform.instance = embracePlatform;
    });

    group('instance', () {
      test('can be overridden with debugEmbraceOverride', () {
        final mockEmbrace = MockEmbrace();
        expect(Embrace.instance, isA<Embrace>());
        expect(Embrace.instance, isNot(mockEmbrace));

        debugEmbraceOverride = mockEmbrace;
        expect(Embrace.instance, mockEmbrace);

        debugEmbraceOverride = null;
        expect(Embrace.instance, isA<Embrace>());
        expect(Embrace.instance, isNot(mockEmbrace));
      });
    });

    group('start', () {
      setUp(() {
        when(
          () => embracePlatform.attachToHostSdk(
            enableIntegrationTesting: any(named: 'enableIntegrationTesting'),
          ),
        ).thenAnswer((_) async => true);
      });

      tearDown(() {
        // The global error handling needs to be reset back to the
        // default value before every test
        if (!belowFlutter_3_1) {
          // ignore: avoid_dynamic_calls
          (PlatformDispatcher.instance as dynamic).onError = null;
        }
      });

      test('calls the supplied action', () async {
        var count = 0;
        await Embrace.instance.start(() => count++);
        expect(count, 1);
      });

      group(
        'when in Flutter 3.1 or above',
        () {
          // We can not throw an uncaught error and verify that it called
          // logDartError because any uncaught errors translates to a
          // failed test
          //
          // So we can only directly test if the global error handling method
          // calls logDartError
          test(
            'handles error caught by dispatcher',
            () async {
              await Embrace.instance.start(() async {
                // ignore: avoid_dynamic_calls
                (PlatformDispatcher.instance as dynamic)
                    .onError('Error', StackTrace.current);
              });
              verify(
                () =>
                    embracePlatform.logDartError(any(), 'Error', any(), any()),
              ).called(1);
            },
          );

          test('does not inject a new error zone', () async {
            final rootErrorZone = Zone.current.errorZone;
            late Zone internalErrorZone;
            await Embrace.instance.start(() async {
              internalErrorZone = Zone.current.errorZone;
            });
            expect(internalErrorZone, rootErrorZone);
          });
        },
        skip: belowFlutter_3_1,
      );

      group(
        'when below Flutter 3.1,',
        () {
          test('Flutter is lower than 3.1', () {
            expect(
              // ignore: avoid_dynamic_calls
              () => (PlatformDispatcher.instance as dynamic).onError,
              throwsNoSuchMethodError,
              reason: 'A version lower than 3.1 is required to run these tests',
            );
          });

          test('catches error', () async {
            await Embrace.instance.start(
              // ignore: only_throw_errors
              () => throw 'Error',
            );
            verify(
              () => embracePlatform.logDartError(any(), 'Error', any(), any()),
            ).called(1);
          });

          test('handles error caught by a zone', () async {
            await Embrace.instance.start(
              () => Zone.current.handleUncaughtError(
                'Error',
                StackTrace.current,
              ),
            );
            verify(
              () => embracePlatform.logDartError(any(), 'Error', any(), any()),
            ).called(1);
          });

          test('injects a new error zone', () async {
            final rootErrorZone = Zone.current.errorZone;
            late Zone internalErrorZone;
            await Embrace.instance.start(() async {
              internalErrorZone = Zone.current.errorZone;
            });
            expect(internalErrorZone, isNot(rootErrorZone));
          });
        },
        skip: !belowFlutter_3_1,
      );

      test('attaches to the host sdk when platform implementation exists', () {
        const enableIntegrationTesting = true;
        Embrace.instance
            .start(() {}, enableIntegrationTesting: enableIntegrationTesting);
        verify(
          () => embracePlatform.attachToHostSdk(
            enableIntegrationTesting: enableIntegrationTesting,
          ),
        ).called(1);
      });
    });

    group('endStartupMoment', () {
      const properties = {'key': 'value'};
      test('logs an error when platform implementation exists', () {
        Embrace.instance.endStartupMoment(properties: properties);
        verify(() => embracePlatform.endAppStartup(properties)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.endAppStartup(properties))
            .thenThrow(MockError());
        Embrace.instance.endStartupMoment(properties: properties);

        verify(
          () => embracePlatform.logInternalError(
            'endStartupMoment',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('logBreadcrumb', () {
      const message = '__message__';
      test('logs an error when platform implementation exists', () {
        Embrace.instance.logBreadcrumb(message);
        verify(() => embracePlatform.logBreadcrumb(message)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.logBreadcrumb(message))
            .thenThrow(MockError());
        Embrace.instance.logBreadcrumb(message);

        verify(
          () => embracePlatform.logInternalError('logBreadcrumb', errorMessage),
        ).called(1);
      });
    });

    group('logInfo', () {
      const message = '__message__';
      const properties = {'key': 'value'};
      test('logs an error when platform implementation exists', () {
        Embrace.instance.logInfo(message, properties: properties);
        verify(() => embracePlatform.logInfo(message, properties)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.logInfo(message, properties))
            .thenThrow(MockError());
        Embrace.instance.logInfo(message, properties: properties);

        verify(() => embracePlatform.logInternalError('logInfo', errorMessage))
            .called(1);
      });
    });

    group('logWarning', () {
      const message = '__message__';
      const properties = {'key': 'value'};
      const allowScreenshots = true;
      test('logs a warning when platform implementation exists', () {
        Embrace.instance.logWarning(
          message,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );
        verify(
          () => embracePlatform.logWarning(
            message,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.logWarning(
            message,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).thenThrow(MockError());
        Embrace.instance.logWarning(
          message,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );

        verify(
          () => embracePlatform.logInternalError(
            'logWarning',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('logError', () {
      const message = '__message__';
      const properties = {'key': 'value'};
      const allowScreenshots = true;
      test('logs an error when platform implementation exists', () {
        Embrace.instance.logError(
          message,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );
        verify(
          () => embracePlatform.logError(
            message,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.logError(
            message,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).thenThrow(MockError());
        Embrace.instance.logError(
          message,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );

        verify(
          () => embracePlatform.logInternalError(
            'logError',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('logNetworkRequest', () {
      const url = '__url__';
      const method = HttpMethod.get;
      const startTime = 123;
      const endTime = 321;
      const bytesSent = 222;
      const bytesReceived = 333;
      const statusCode = 200;
      const error = '__error__';
      const traceId = '__traceId__';
      test(
          'logs the specified network request when '
          'platform implementation exists', () {
        Embrace.instance.logNetworkRequest(
          url: url,
          method: method,
          startTime: startTime,
          endTime: endTime,
          bytesSent: bytesSent,
          bytesReceived: bytesReceived,
          statusCode: statusCode,
          error: error,
          traceId: traceId,
        );
        verify(
          () => embracePlatform.logNetworkRequest(
            url: url,
            method: method,
            startTime: startTime,
            endTime: endTime,
            bytesSent: bytesSent,
            bytesReceived: bytesReceived,
            statusCode: statusCode,
            error: error,
            traceId: traceId,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.logNetworkRequest(
            url: url,
            method: method,
            startTime: startTime,
            endTime: endTime,
            bytesSent: bytesSent,
            bytesReceived: bytesReceived,
            statusCode: statusCode,
            error: error,
            traceId: traceId,
          ),
        ).thenThrow(MockError());
        Embrace.instance.logNetworkRequest(
          url: url,
          method: method,
          startTime: startTime,
          endTime: endTime,
          bytesSent: bytesSent,
          bytesReceived: bytesReceived,
          statusCode: statusCode,
          error: error,
          traceId: traceId,
        );

        verify(
          () => embracePlatform.logInternalError(
            'logNetworkRequest',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('startMoment', () {
      const momentName = '__momentName__';
      const identifier = '__identifier__';
      const properties = {'key': 'value'};
      const allowScreenshots = true;
      test(
          'starts the specified moment when '
          'platform implementation exists', () {
        Embrace.instance.startMoment(
          momentName,
          identifier: identifier,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );
        verify(
          () => embracePlatform.startMoment(
            momentName,
            identifier,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.startMoment(
            momentName,
            identifier,
            properties,
            allowScreenshot: allowScreenshots,
          ),
        ).thenThrow(MockError());
        Embrace.instance.startMoment(
          momentName,
          identifier: identifier,
          properties: properties,
          allowScreenshot: allowScreenshots,
        );

        verify(
          () => embracePlatform.logInternalError(
            'startMoment',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('endMoment', () {
      const momentName = '__momentName__';
      const identifier = '__identifier__';
      const properties = {'key': 'value'};
      test('ends the specified moment when platform implementation exists', () {
        Embrace.instance.endMoment(
          momentName,
          identifier: identifier,
          properties: properties,
        );
        verify(
          () => embracePlatform.endMoment(
            momentName,
            identifier,
            properties,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.endMoment(
            momentName,
            identifier,
            properties,
          ),
        ).thenThrow(MockError());
        Embrace.instance.endMoment(
          momentName,
          identifier: identifier,
          properties: properties,
        );

        verify(
          () => embracePlatform.logInternalError(
            'endMoment',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('startView', () {
      const view = '__view__';
      test('starts the specified view when platform implementation exists', () {
        Embrace.instance.startView(view);
        verify(() => embracePlatform.startView(view)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.startView(view)).thenThrow(MockError());
        Embrace.instance.startView(view);

        verify(
          () => embracePlatform.logInternalError(
            'startView',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('endView', () {
      const view = '__view__';
      test('ends the specified view when platform implementation exists', () {
        Embrace.instance.endView(view);
        verify(() => embracePlatform.endView(view)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.endView(view)).thenThrow(MockError());
        Embrace.instance.endView(view);

        verify(
          () => embracePlatform.logInternalError(
            'endView',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserIdentifier', () {
      const userIdentifier = '__userIdentifier__';
      test('sets the user name when platform implementation exists', () {
        Embrace.instance.setUserIdentifier(userIdentifier);
        verify(() => embracePlatform.setUserIdentifier(userIdentifier))
            .called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.setUserIdentifier(userIdentifier))
            .thenThrow(MockError());
        Embrace.instance.setUserIdentifier(userIdentifier);

        verify(
          () => embracePlatform.logInternalError(
            'setUserIdentifier',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserIdentifier', () {
      test(
          'clears the user identifier when '
          'platform implementation exists', () {
        Embrace.instance.clearUserIdentifier();
        verify(embracePlatform.clearUserIdentifier).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearUserIdentifier).thenThrow(MockError());
        Embrace.instance.clearUserIdentifier();

        verify(
          () => embracePlatform.logInternalError(
            'clearUserIdentifier',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserName', () {
      const userName = '__userName__';
      test('sets the user name when platform implementation exists', () {
        Embrace.instance.setUserName(userName);
        verify(() => embracePlatform.setUserName(userName)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.setUserName(userName))
            .thenThrow(MockError());
        Embrace.instance.setUserName(userName);

        verify(
          () => embracePlatform.logInternalError(
            'setUserName',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserName', () {
      test('sets the user email when platform implementation exists', () {
        Embrace.instance.clearUserName();
        verify(embracePlatform.clearUserName).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearUserName).thenThrow(MockError());
        Embrace.instance.clearUserName();

        verify(
          () => embracePlatform.logInternalError(
            'clearUserName',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserEmail', () {
      const email = '__email__';
      test('sets the user email when platform implementation exists', () {
        Embrace.instance.setUserEmail(email);
        verify(() => embracePlatform.setUserEmail(email)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.setUserEmail(email)).thenThrow(MockError());
        Embrace.instance.setUserEmail(email);

        verify(
          () => embracePlatform.logInternalError(
            'setUserEmail',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserEmail', () {
      test('clears the user email when platform implementation exists', () {
        Embrace.instance.clearUserEmail();
        verify(embracePlatform.clearUserEmail).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearUserEmail).thenThrow(MockError());
        Embrace.instance.clearUserEmail();

        verify(
          () => embracePlatform.logInternalError(
            'clearUserEmail',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserAsPayer', () {
      test('sets the user as payer when platform implementation exists', () {
        Embrace.instance.setUserAsPayer();
        verify(embracePlatform.setUserAsPayer).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.setUserAsPayer).thenThrow(MockError());
        Embrace.instance.setUserAsPayer();

        verify(
          () => embracePlatform.logInternalError(
            'setUserAsPayer',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserAsPayer', () {
      test('clears the user as payer when platform implementation exists', () {
        Embrace.instance.clearUserAsPayer();
        verify(embracePlatform.clearUserAsPayer).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearUserAsPayer).thenThrow(MockError());
        Embrace.instance.clearUserAsPayer();

        verify(
          () => embracePlatform.logInternalError(
            'clearUserAsPayer',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('setUserPersona', () {
      const persona = '__persona__';
      test('sets the user persona when platform implementation exists', () {
        Embrace.instance.setUserPersona(persona);
        verify(() => embracePlatform.setUserPersona(persona)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.setUserPersona(persona))
            .thenThrow(MockError());
        Embrace.instance.setUserPersona(persona);

        verify(
          () => embracePlatform.logInternalError(
            'setUserPersona',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearUserPersona', () {
      const persona = '__persona__';
      test('clears the user persona when platform implementation exists', () {
        Embrace.instance.clearUserPersona(persona);
        verify(() => embracePlatform.clearUserPersona(persona)).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.clearUserPersona(persona))
            .thenThrow(MockError());
        Embrace.instance.clearUserPersona(persona);

        verify(
          () => embracePlatform.logInternalError(
            'clearUserPersona',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('clearAllUserPersonas', () {
      test('clears user personas when platform implementation exists', () {
        Embrace.instance.clearAllUserPersonas();
        verify(embracePlatform.clearAllUserPersonas).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(embracePlatform.clearAllUserPersonas).thenThrow(MockError());
        Embrace.instance.clearAllUserPersonas();

        verify(
          () => embracePlatform.logInternalError(
            'clearAllUserPersonas',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('addSessionProperty', () {
      const key = '__key__';
      const value = '__value__';
      const permanent = false;
      test('adds session property when platform implementation exists', () {
        Embrace.instance.addSessionProperty(key, value);
        verify(
          () => embracePlatform.addSessionProperty(
            key,
            value,
            permanent: permanent,
          ),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(
          () => embracePlatform.addSessionProperty(
            key,
            value,
            permanent: permanent,
          ),
        ).thenThrow(MockError());
        Embrace.instance.addSessionProperty(key, value);

        verify(
          () => embracePlatform.logInternalError(
            'addSessionProperty',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('removeSessionProperty', () {
      const key = '__key__';
      test('removes session property when platform implementation exists', () {
        Embrace.instance.removeSessionProperty(key);
        verify(
          () => embracePlatform.removeSessionProperty(key),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.removeSessionProperty(key))
            .thenThrow(MockError());
        Embrace.instance.removeSessionProperty(key);

        verify(
          () => embracePlatform.logInternalError(
            'removeSessionProperty',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('getSessionProperties', () {
      test('getSessionProperties when platform implementation exists', () {
        Embrace.instance.getSessionProperties();
        verify(
          () => embracePlatform.getSessionProperties(),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.getSessionProperties())
            .thenThrow(MockError());
        final properties = Embrace.instance.getSessionProperties();

        expect(properties, isA<Future<Map<String, String>>>());

        verify(
          () => embracePlatform.logInternalError(
            'getSessionProperties',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('endSession', () {
      test('ends session when platform implementation exists', () {
        Embrace.instance.endSession();
        verify(
          () => embracePlatform.endSession(
            clearUserInfo: any(named: 'clearUserInfo'),
          ),
        ).called(1);
      });

      test('clearUserInfo defaults to true', () {
        Embrace.instance.endSession();
        verify(
          // ignore: avoid_redundant_argument_values
          () => embracePlatform.endSession(clearUserInfo: true),
        ).called(1);
      });

      test(
          'logs internal error when platform implementation '
          'throws an error', () {
        when(() => embracePlatform.endSession()).thenThrow(MockError());
        Embrace.instance.endSession();

        verify(
          () => embracePlatform.logInternalError(
            'endSession',
            errorMessage,
          ),
        ).called(1);
      });
    });

    group('getDeviceId', () {
      test('returns device ID when platform implementation exists', () async {
        const platformName = '__test_platform__';
        when(embracePlatform.getDeviceId).thenAnswer((_) async => platformName);

        final actualPlatformName = await Embrace.instance.getDeviceId();
        expect(actualPlatformName, equals(platformName));
      });
    });
  });
}
